from datetime import datetime
from typing import List

import orjson
from fastapi import APIRouter, Depends, HTTPException, status
from fastapi.responses import ORJSONResponse
from sqlalchemy import orm
from starlette.responses import Response

from balsam import schemas
from balsam.schemas import MAX_ITEMS_PER_BULK_OP
from balsam.server import ValidationError
from balsam.server.auth import get_auth_method, get_webuser_session
from balsam.server.models import Job, crud
from balsam.server.pubsub import pubsub
from balsam.server.utils import Paginator

from .filters import JobQuery

router = APIRouter()
auth = get_auth_method()


@router.get("/", response_class=Response)
def list(
    db: orm.Session = Depends(get_webuser_session),
    user: schemas.UserOut = Depends(auth),
    paginator: Paginator[Job] = Depends(Paginator),
    q: JobQuery = Depends(JobQuery),
) -> Response:
    """List the user's Jobs."""
    count, jobs = crud.jobs.fetch(db, owner=user, paginator=paginator, filterset=q)
    return Response(content=orjson.dumps({"count": count, "results": jobs}), media_type="application/json")


@router.get("/{job_id}", response_class=ORJSONResponse)
def read(
    job_id: int, db: orm.Session = Depends(get_webuser_session), user: schemas.UserOut = Depends(auth)
) -> ORJSONResponse:
    """Get a Job by id."""
    count, jobs = crud.jobs.fetch(db, owner=user, job_id=job_id)
    return ORJSONResponse(content=jobs[0])


@router.post("/", response_class=ORJSONResponse, status_code=status.HTTP_201_CREATED)
def bulk_create(
    jobs: List[schemas.ServerJobCreate],
    db: orm.Session = Depends(get_webuser_session),
    user: schemas.UserOut = Depends(auth),
) -> ORJSONResponse:
    """Create a list of Jobs."""
    if len(jobs) > MAX_ITEMS_PER_BULK_OP:
        raise HTTPException(
            status_code=400, detail=f"Cannot bulk-create more than {MAX_ITEMS_PER_BULK_OP} in a single API call."
        )
    new_jobs = crud.jobs.bulk_create(db, owner=user, job_specs=jobs)
    db.commit()
    # TODO: Pubsub.publish using jsonable_encoder: killing performance for many jobs
    # If re-integrating pubsub, need to root out fastapi.jsonable_encoder.
    return ORJSONResponse(content=new_jobs, status_code=status.HTTP_201_CREATED)


@router.patch("/")
def bulk_update(
    jobs: List[schemas.JobBulkUpdate],
    db: orm.Session = Depends(get_webuser_session),
    user: schemas.UserOut = Depends(auth),
) -> int:
    """Update a list of Jobs"""
    if len(jobs) > MAX_ITEMS_PER_BULK_OP:
        raise HTTPException(
            status_code=400, detail=f"Cannot bulk-update more than {MAX_ITEMS_PER_BULK_OP} in a single API call."
        )
    now = datetime.utcnow()
    patch_dicts = {job.id: {**job.dict(exclude_unset=True, exclude={"id"}), "last_update": now} for job in jobs}
    if len(jobs) > len(patch_dicts):
        raise ValidationError("Duplicate Job ID keys provided")
    num_updated = crud.jobs.bulk_update(db, owner=user, patch_dicts=patch_dicts)
    db.commit()
    return num_updated


@router.put("/")
def query_update(
    update_fields: schemas.JobUpdate,
    db: orm.Session = Depends(get_webuser_session),
    user: schemas.UserOut = Depends(auth),
    q: JobQuery = Depends(JobQuery),
) -> int:
    """Apply the same update to all Jobs selected by the query."""
    data = update_fields.dict(exclude_unset=True)
    data["last_update"] = datetime.utcnow()
    num_updated = crud.jobs.update_query(db, owner=user, update_data=data, filterset=q)
    db.commit()
    return num_updated


@router.put("/{job_id}")
def update(
    job_id: int,
    job: schemas.JobUpdate,
    db: orm.Session = Depends(get_webuser_session),
    user: schemas.UserOut = Depends(auth),
) -> int:
    """Update a Job by id."""
    data = job.dict(exclude_unset=True)
    data["last_update"] = datetime.utcnow()
    patch = {job_id: data}
    num_updated = crud.jobs.bulk_update(db, owner=user, patch_dicts=patch)
    db.commit()
    return num_updated


@router.delete("/{job_id}", status_code=status.HTTP_204_NO_CONTENT)
def delete(
    job_id: int, db: orm.Session = Depends(get_webuser_session), user: schemas.UserOut = Depends(auth)
) -> None:
    """Delete a Job by id."""
    crud.jobs.delete_query(db, owner=user, job_id=job_id)
    db.commit()
    pubsub.publish(user.id, "bulk-delete", "job", {"ids": [job_id]})


@router.delete("/", status_code=status.HTTP_204_NO_CONTENT)
def query_delete(
    db: orm.Session = Depends(get_webuser_session),
    user: schemas.UserOut = Depends(auth),
    q: JobQuery = Depends(JobQuery),
) -> int:
    """Delete all jobs selected by the query."""
    num_deleted = crud.jobs.delete_query(db, owner=user, filterset=q)
    db.commit()
    return num_deleted



from typing import Callable, Iterator

from fastapi import Depends
from sqlalchemy import orm

from balsam import schemas
from balsam.server import settings
from balsam.server.models import get_session

from .token import user_from_token

AUTH_METHODS = {
    "user_from_token": user_from_token,
}


def get_auth_method() -> Callable[..., schemas.UserOut]:
    return AUTH_METHODS[settings.auth.auth_method]


def get_admin_session() -> Iterator[orm.Session]:
    """
    Use with FastAPI Depends() to get a `postgres` superuser
    database connection.
    """
    session = get_session(user=None)
    try:
        yield session
    except:  # noqa: E722
        session.rollback()
        raise
    finally:
        session.close()


def get_webuser_session(user: schemas.UserOut = Depends(get_auth_method())) -> Iterator[orm.Session]:
    """
    Use with FastAPI Depends() to get a user-scoped database connection automatically from
    the request credentials.
    """
    session = get_session(user=user)
    try:
        yield session
    except:  # noqa: E722
        session.rollback()
        raise
    finally:
        session.close()



from datetime import datetime
from typing import Any, Dict

from fastapi import APIRouter, Depends, status
from sqlalchemy import orm

from balsam import schemas
from balsam.server.auth import get_auth_method, get_webuser_session
from balsam.server.models import Site, crud
from balsam.server.pubsub import pubsub
from balsam.server.utils import Paginator

from .filters import SiteQuery

router = APIRouter()
auth = get_auth_method()
print("auth is", auth)


@router.get("/", response_model=schemas.PaginatedSitesOut)
def list(
    q: SiteQuery = Depends(SiteQuery),
    db: orm.Session = Depends(get_webuser_session),
    user: schemas.UserOut = Depends(auth),
    paginator: Paginator[Site] = Depends(Paginator),
) -> Dict[str, Any]:
    """List Sites belonging to the user."""
    count, sites = crud.sites.fetch(
        db,
        owner=user,
        paginator=paginator,
        filterset=q,
    )
    return {"count": count, "results": sites}


@router.get("/{site_id}", response_model=schemas.SiteOut)
def read(site_id: int, db: orm.Session = Depends(get_webuser_session), user: schemas.UserOut = Depends(auth)) -> Site:
    """Fetch a Sites by id."""
    _, site = crud.sites.fetch(db, owner=user, site_id=site_id)
    assert isinstance(site, Site)
    return site


@router.post("/", response_model=schemas.SiteOut, status_code=status.HTTP_201_CREATED)
def create(
    site: schemas.SiteCreate,
    db: orm.Session = Depends(get_webuser_session),
    user: schemas.UserOut = Depends(auth),
) -> schemas.SiteOut:
    """Create a new Balsam Site."""
    new_site = crud.sites.create(db, owner=user, site=site)
    result = schemas.SiteOut.from_orm(new_site)
    db.commit()
    pubsub.publish(user.id, "create", "site", result)
    return result


@router.put("/{site_id}", response_model=schemas.SiteOut)
def update(
    site_id: int,
    site: schemas.SiteUpdate,
    db: orm.Session = Depends(get_webuser_session),
    user: schemas.UserOut = Depends(auth),
) -> Site:
    """Update a Balsam Site by id."""
    data = site.dict(exclude_unset=True)
    data["last_refresh"] = datetime.utcnow()
    updated_site = crud.sites.update(db, owner=user, site_id=site_id, update_data=data)
    data["id"] = site_id
    db.commit()
    pubsub.publish(user.id, "update", "site", data)
    return updated_site


@router.delete("/{site_id}", status_code=status.HTTP_204_NO_CONTENT)
def delete(
    site_id: int, db: orm.Session = Depends(get_webuser_session), user: schemas.UserOut = Depends(auth)
) -> None:
    """Delete a Balsam Site by id. All associated Apps, Jobs, and BatchJobs are destroyed."""
    crud.sites.delete(db, owner=user, site_id=site_id)
    db.commit()
    pubsub.publish(user.id, "delete", "site", {"id": site_id})



from typing import Generic, TypeVar

from fastapi import Query as APIQuery
from sqlalchemy.orm import Query as SQLQuery
from sqlalchemy.sql import Select

from balsam.schemas import MAX_PAGE_SIZE

T = TypeVar("T")


class Paginator(Generic[T]):
    def __init__(
        self,
        limit: int = APIQuery(MAX_PAGE_SIZE, le=MAX_PAGE_SIZE, description="Maximum number of items to return."),
        offset: int = APIQuery(0, ge=0, description="Starting index from which to retrieve results."),
    ):
        self.limit = limit
        self.offset = offset

    def paginate(self, iterable: "SQLQuery[T]") -> "SQLQuery[T]":
        return iterable[self.offset : self.offset + self.limit]  # type: ignore

    def paginate_core(self, stmt: "Select") -> "Select":
        if self.limit is not None:
            stmt = stmt.limit(self.limit)
        return stmt.offset(self.offset)

    def __new__(
        cls,
        limit: int = APIQuery(MAX_PAGE_SIZE, le=MAX_PAGE_SIZE, description="Maximum number of items to return."),
        offset: int = APIQuery(0, ge=0, description="Starting index from which to list items."),
    ) -> "Paginator[T]":
        return super().__new__(cls)  # type: ignore



import logging
from datetime import datetime, timedelta
from typing import Any, Dict, List, Tuple

from sqlalchemy import Float, func, orm
from sqlalchemy.sql import Select, cast, select, update

from balsam import schemas
from balsam.server import ValidationError, models
from balsam.server.routers.filters import SessionQuery

from .jobs import do_update_jobs, owned_job_selector, select_jobs_for_update

logger = logging.getLogger(__name__)

SESSION_EXPIRE_PERIOD = timedelta(minutes=5)
SESSION_SWEEP_PERIOD = timedelta(minutes=1)
latest_sweep_time: Dict[int, datetime] = {}
Session = orm.Session
Query = orm.Query


def owned_session_query(db: Session, owner: schemas.UserOut) -> "Query[models.Session]":
    return db.query(models.Session).join(models.Site).filter(models.Site.owner_id == owner.id)  # type: ignore


def fetch(db: Session, owner: schemas.UserOut, filterset: SessionQuery) -> Tuple[int, List[models.Session]]:
    qs = owned_session_query(db, owner)
    result = filterset.apply_filters(qs).order_by(models.Session.id).all()
    count = len(result)
    return count, result


def _timeout_jobs(db: Session, session: models.Session) -> None:
    zombies = (
        select((models.Job.__table__,))
        .where(models.Job.session_id == session.id)
        .where(models.Job.state == "RUNNING")
    )
    update_jobs, transfer_items_by_jobid = select_jobs_for_update(db, zombies)
    timeout_dat = {
        "state": "RUN_TIMEOUT",
        "state_data": {"message": "Session expired: job was stuck in RUNNING state"},
    }
    if update_jobs:
        do_update_jobs(
            db,
            update_jobs,
            transfer_items_by_jobid,
            patch_dicts={job.id: timeout_dat.copy() for job in update_jobs},
        )
        logger.info(f"Timed out {len(update_jobs)} running jobs in expired session {session.id}")


def _clear_stale_sessions(db: Session, owner: schemas.UserOut) -> None:
    now = datetime.utcnow()

    last_sweep = latest_sweep_time.get(owner.id)
    if last_sweep is not None and now < last_sweep + SESSION_SWEEP_PERIOD:
        return

    latest_sweep_time[owner.id] = now
    expiry_time = now - SESSION_EXPIRE_PERIOD

    sess_ids = []

    expired_sessions = owned_session_query(db, owner).filter(models.Session.heartbeat <= expiry_time).all()
    for sess in expired_sessions:
        logger.info(f"Session {sess.id} expired: last heartbeat was {sess.heartbeat}; expiry_time is {expiry_time}")
        sess_ids.append(sess.id)

    finished_launcher_sessions = (
        owned_session_query(db, owner)  # type: ignore
        .join(models.BatchJob, models.BatchJob.id == models.Session.batch_job_id)
        .filter(models.BatchJob.state == "finished")
        .all()
    )
    for sess in finished_launcher_sessions:
        if sess.id not in sess_ids:
            logger.info(f"Session {sess.id} expired: the assosciated BatchJob state is finished.")
            expired_sessions.append(sess)
            sess_ids.append(sess.id)

    for session in expired_sessions:
        _timeout_jobs(db, session)

    db.query(models.Session).filter(models.Session.id.in_(sess_ids)).delete(synchronize_session=False)
    db.flush()


def create(db: Session, owner: schemas.UserOut, session: schemas.SessionCreate) -> models.Session:
    _clear_stale_sessions(db, owner)

    site_id = (
        db.query(models.Site.id).filter(models.Site.owner_id == owner.id, models.Site.id == session.site_id).scalar()  # type: ignore
    )
    if site_id is None:
        raise ValidationError(f"No site with ID {session.site_id}")
    if session.batch_job_id is not None:
        batch_job_id = (
            db.query(models.BatchJob.id)
            .filter(models.BatchJob.site_id == site_id)
            .filter(models.BatchJob.id == session.batch_job_id)
        ).scalar()  # type: ignore
        if batch_job_id is None:
            raise ValidationError(f"No batch_job with id {session.batch_job_id}")

    created_session = models.Session(**session.dict())
    db.add(created_session)
    db.flush()
    return created_session


def _acquire_jobs(db: orm.Session, job_q: Select, session: models.Session) -> List[Dict[str, Any]]:
    acquired_jobs = [{str(key): value for key, value in job.items()} for job in db.execute(job_q).mappings()]
    acquired_ids = [job["id"] for job in acquired_jobs]

    stmt = update(models.Job.__table__).where(models.Job.id.in_(acquired_ids)).values(session_id=session.id)

    # Do not overwrite job.batch_job_id with a Session that has no batch_job_id:
    if session.batch_job_id is not None:
        stmt = stmt.values(batch_job_id=session.batch_job_id)
        for job in acquired_jobs:
            job["batch_job_id"] = session.batch_job_id
            job["session_id"] = session.id

    db.execute(stmt)
    db.flush()
    logger.debug(f"Acquired {len(acquired_jobs)} jobs")
    return acquired_jobs


def _footprint_func() -> Any:
    footprint = cast(models.Job.num_nodes, Float) / cast(models.Job.node_packing_count, Float)
    return (
        func.sum(footprint)
        .over(
            order_by=(
                models.Job.num_nodes.asc(),
                models.Job.node_packing_count.desc(),
                models.Job.wall_time_min.desc(),
                models.Job.id.asc(),
            )
        )
        .label("aggregate_footprint")
    )


def acquire(
    db: Session, owner: schemas.UserOut, session_id: int, spec: schemas.SessionAcquire
) -> List[Dict[str, Any]]:

    session = (owned_session_query(db, owner).filter(models.Session.id == session_id)).one()
    session.heartbeat = datetime.utcnow()

    logger.debug(f"Acquire: filtering for jobs with states: {spec.states}")

    # Select unlocked jobs at this Site matching the state / tags criteria
    job_q = (
        owned_job_selector(owner)
        .where(models.App.site_id == session.site_id)
        .where(models.Job.session_id.is_(None))  # type: ignore
        .where(models.Job.state.in_(spec.states))
        .where(models.Job.tags.contains(spec.filter_tags))  # type: ignore
    )

    if spec.app_ids:
        job_q = job_q.where(models.Job.app_id.in_(spec.app_ids))

    if spec.max_wall_time_min:
        job_q = job_q.where(models.Job.wall_time_min <= spec.max_wall_time_min)  # By time

    if spec.max_nodes_per_job:
        job_q = job_q.where(models.Job.num_nodes <= spec.max_nodes_per_job)

    if spec.serial_only:
        job_q = job_q.where(models.Job.num_nodes == 1).where(models.Job.ranks_per_node == 1)

    # This is the branch taken by Processing and Serial Mode Launcher:
    if spec.max_aggregate_nodes is None:
        # NO footprint calculation; NO ordering is needed!
        job_q = job_q.limit(spec.max_num_jobs).with_for_update(of=models.Job.__table__, skip_locked=True)
        return _acquire_jobs(db, job_q, session)

    # MPI Mode Launcher will take this path:
    lock_ids_q = (
        job_q.with_only_columns([models.Job.id])
        .order_by(
            models.Job.num_nodes.asc(),
            models.Job.node_packing_count.desc(),
            models.Job.wall_time_min.desc(),
        )
        .limit(spec.max_num_jobs)
        .with_for_update(of=models.Job.__table__, skip_locked=True)
    )
    locked_ids = db.execute(lock_ids_q).scalars().all()

    subq = select(models.Job.__table__, _footprint_func()).where(models.Job.id.in_(locked_ids)).subquery()  # type: ignore
    cols = [c for c in subq.c if c.name not in ["aggregate_footprint", "session_id"]]
    job_q = select(cols).where(subq.c.aggregate_footprint <= spec.max_aggregate_nodes)
    return _acquire_jobs(db, job_q, session)


def tick(db: Session, owner: schemas.UserOut, session_id: int) -> datetime:
    in_db = owned_session_query(db, owner).filter(models.Session.id == session_id).one()
    ts = datetime.utcnow()
    in_db.heartbeat = ts
    db.flush()

    # Clear after updating heartbeat, to avoid clearing self
    _clear_stale_sessions(db, owner)
    return ts


def delete(db: Session, owner: schemas.UserOut, session_id: int) -> None:
    qs = owned_session_query(db, owner).filter(models.Session.id == session_id)
    session = qs.one()

    _timeout_jobs(db, session)

    db.query(models.Session).filter(models.Session.id == session_id).delete(synchronize_session=False)
    db.flush()
